# Generated by Django 3.1.4 on 2021-03-14 20:05

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import ryzom_django_channels.pubsub


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Room',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
            ],
            bases=(ryzom_django_channels.pubsub.Publishable, models.Model),
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('message', models.CharField(max_length=1024)),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('room', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ryzom_django_channels_example.room')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL)),
            ],
            bases=(ryzom_django_channels.pubsub.Publishable, models.Model),
        ),
    ]



from ryzom_django_channels.models import (
    Publication,
    Subscription,
    Registration
)


model_templates = dict()


def model_template(name):
    global model_templates
    def decorator(component):
        model_templates[name] = component
        return component
    return decorator


class ReactiveBase:
    view = None

    def to_html(self, *content, **context):
        self.reactive_setup(**context)
        return super(ReactiveBase, self).to_html(*content, **context)

    def reactive_setup(self, **context):
        self.view = self.get_view(**context)

        if self.view is None:
            parent = self.parent or self
            while parent and parent.parent:
                if hasattr(parent, 'view'):
                    break
                parent = parent.parent
            try:
                self.view = parent.view
            except AttributeError:
                raise AttributeError('The current view cannot be found')

        if not hasattr(self.view, 'client'):
            raise AttributeError(
                'The current view has no attribute "client".'
                ' Maybe you forgot to call view.get_token()'
                ' in your main component?')

    def get_view(self, **context):
        if 'view' in context:
            return context['view']


class SubscribeComponentMixin(ReactiveBase):
    @property
    def model_template(self):
        raise AttributeError(
            f'{self} is missing attribute "model_template"'
        )

    def reactive_setup(self, **context):
        if not hasattr(self, 'subscribe_options'):
            self.subscribe_options = {}

        super().reactive_setup(**context)

        if hasattr(self, 'publication'):
            self.create_subscription()


    def create_subscription(self):
        publication = Publication.objects.get(name=self.publication)
        subscription = Subscription.objects.create(
            client=self.view.client,
            publication=publication,
            subscriber_id=self.id,
            subscriber_module=self.__module__,
            subscriber_class=self.__class__.__name__,
            options=self.subscribe_options,
        )

        self.get_content(publication, subscription)

    def get_content(self, publication, subscription):
        template = model_templates[self.model_template]

        content = []
        for obj in subscription.get_queryset():
            content.append(template(obj))

        self.content = content

    @classmethod
    def get_queryset(self, qs, opts):
        return qs


class ReactiveComponentMixin(ReactiveBase):
    register = None

    def reactive_setup(self, **context):
        super().reactive_setup(**context)

        if hasattr(self, 'register'):
            self.create_registration()

    def create_registration(self):
        existent = Registration.objects.filter(
            name=self.get_register(),
            client=self.view.client
        ).first()

        if existent:
            existent.subscriber_id = self.id
            existent.subscriber_parent = self.parent.id
            existent.save()

        else:
            Registration.objects.create(
                name=self.get_register(),
                client=self.view.client,
                subscriber_id=self.id,
                subscriber_parent=self.parent.id,
            )

    def get_register(self):
        if self.register is None:
            raise AttributeError(f'{self}.register is not defined')

        return self.register



from ryzom.components import CList, Component, CTree, HTMLPayload, Markdown, Text
from py2js.transpiler import transpile_body

templates = dict()

def template(name, *wrappers):
    global templates
    def decorator(component):
        if wrappers:
            templates[name] = CTree(*wrappers + (component,))
        else:
            templates[name] = component
        return component
    return decorator


BASIC_TAGS = (
    'a', 'abbr', 'acronym', 'address', 'applet', 'article', 'aside',
    'audio', 'b', 'basefont', 'bdi', 'bdo', 'bgsound', 'big', 'blink',
    'blockquote', 'body', 'button', 'canvas', 'caption', 'center',
    'cite', 'code', 'colgroup', 'content', 'contributors.txt', 'data',
    'datalist', 'dd', 'del', 'details', 'dfn', 'dialog', 'dir', 'div', 'dl',
    'dt', 'em', 'fieldset', 'figcaption', 'figure', 'font', 'footer',
    'frame', 'frameset', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'head',
    'header', 'Heading_Elements', 'hgroup', 'i', 'iframe', 'image',
    'ins', 'isindex', 'kbd', 'keygen', 'label', 'legend', 'li', 'listing',
    'main', 'map', 'mark', 'marquee', 'menu', 'menu#context_menu', 'menuitem',
    'meter', 'multicol', 'nav', 'nextid', 'nobr', 'noembed', 'noframes',
    'noscript', 'object', 'ol', 'optgroup', 'option', 'output', 'p', 'picture',
    'plaintext', 'polygon', 'portal', 'pre', 'progress', 'q', 'rb', 'rp', 'rt',
    'rtc', 'ruby', 's', 'samp', 'section', 'select', 'shadow', 'slot', 'small',
    'spacer', 'span', 'strike', 'strong', 'sub', 'summary', 'sup', 'svg',
    'table', 'tbody', 'td', 'template', 'textarea', 'tfoot', 'th', 'thead',
    'time', 'title', 'tr', 'tt', 'u', 'ul', 'var', 'video', 'xmp',
)

for tag in BASIC_TAGS:
    locals()[tag.capitalize()] = type(tag.capitalize(), (Component,), {'tag': tag})

SELFCLOSE_TAGS = (
    'area', 'base', 'br', 'col', 'embed', 'hr', 'img', 'link', 'meta',
    'param', 'source', 'track', 'wbr',
)

for tag in SELFCLOSE_TAGS:
    locals()[tag.capitalize()] = type(
        tag.capitalize(),
        (Component,),
        {'tag': tag, 'selfclose': True},
    )


class Form(Component):
    tag = 'form'
    attrs = {'enctype': 'multipart/form-data'}


class Script(Component):
    tag = 'script'
    attrs = {'type': 'text/javascript'}


class Style(Component):
    tag = 'style'
    attrs = {'type': 'text/css'}


class Stylesheet(Link):
    attrs = {'type': 'text/css', 'rel': 'stylesheet'}


class Input(Component):
    tag = 'input'

    def __init__(self, *content, **attrs):
        if 'name' in attrs:
            attrs.setdefault('input_id', f'id_{attrs["name"]}')
            attrs.setdefault('label_id', f'id_{attrs["name"]}_label')
        super().__init__(*content, **attrs)


class Icon(Component):
    tag = 'i'


class Head(Component):
    def __init__(self, *content, **attrs):
        super().__init__(*content, **attrs)
        self.content += [
            Meta(
                name='viewport',
                content='width=device-width, initial-scale=1.0',
            ),
            Meta(charset='utf-8'),
        ]


class Html(Component):
    tag = 'html'
    body_class = Body
    head_class = Head

    def __init__(self, *content, **context):
        if 'head' not in context:
            context['head'] = self.head_class(*context.pop('extra_head', []))

        if 'body' not in context:
            context['body'] = self.body_class(*content)

        super().__init__(**context)

        for src in self.stylesheets:
            if not src:
                continue

            if hasattr(src, 'to_html'):
                self.head.content.append(src)
            else:
                self.head.content.append(Stylesheet(href=src))

        for src in self.scripts:
            if not src:
                continue

            if hasattr(src, 'to_html'):
                self.head.content.append(src)
            elif callable(src):
                self.head.content.append(
                    Script(transpile_body(src))
                )
            else:
                self.head.content.append(Script(src=src))

        if title := getattr(self, 'title', None):
            self.__dict__['title'] = Title(title)
            self.head.addchild(self.title)



from django import forms
from django.urls import path, reverse
from django.views import generic

from ryzom_django.html import template
from ryzom_django_mdc import html


class ExampleDocument(html.Html):
    title = 'Secure Elections with homomorphic encryption'


# Serves to demonstrate template composition based on multi level nesting
class ExampleCard(html.Div):
    def __init__(self, *content, **context):
        super().__init__(*content, style='max-width: 20em; margin: auto')


# in general, you wouldn't go through a template_name, but since this is useful
# to design views that you import from external packages, we have this example
# here, it also shows how you can compose by nesting different layout objects
@html.template('form.html', ExampleDocument, ExampleCard)
class ExampleFormViewComponent(html.Div):
    title = 'Example form view'

    def to_html(self, view, form, **context):
        # view and form come from the default context, we're spreading them as
        # nice, required variables for this template.

        # As you can imagine, having templates in Python not only gives you all
        # the programing power, but it also allows you to use a debugger
        # breakpoint() which was not possible with traditionnal templates.

        content = []

        if view.request.method == 'POST' and form.is_valid():
            content += [
                html.Div(
                    html.H3('Form post success!'),
                    html.MDCList(*[
                        html.MDCListItem(f'{key}: {value}')
                        for key, value in form.cleaned_data.items()
                    ])
                )
            ]

        content += [
            html.Form(
                html.CSRFInput(view.request),
                form,
                html.MDCButton(form.submit_label),
                method='POST',
            )
        ]

        return super().to_html(*content, **context)


class ExampleForm(forms.Form):
    submit_label = 'Send'

    char = forms.CharField(
        help_text='Non required example text input',
    )
    email = forms.EmailField(
        help_text='Valid email required',
    )
    boolean = forms.BooleanField(
        help_text='Required boolean!',
    )
    checkboxes = forms.MultipleChoiceField(
        choices=(('a', 'a'), ('b', 'b')),
        widget=forms.CheckboxSelectMultiple,
        help_text='Required checkbox multiple',
    )
    datetime = forms.SplitDateTimeField(
        widget=forms.SplitDateTimeWidget(
            date_attrs=dict(type='date'),
            time_attrs=dict(type='time'),
        ),
        help_text='Required help text',
    )
    textarea = forms.CharField(
        widget=forms.Textarea,
        help_text='This is the help text'
    )

    document = forms.FileField(
        widget=forms.FileInput,
        help_text='Choose a file'
    )

    select = forms.ChoiceField(
        widget=forms.Select,
        choices=(
            ('Test', (
                (1, 'the first one'),
                (2, 'the second'),
                (3, 'the thirf')
            )),
            ('Re', (
                (4, '444'),
                (5, '555')
            )),
        ),
        initial=5
    )


# Finally, a Django FormView, there's nothing to see here because of how well
# Ryzom integrates with Django. Of course you're free to make views that do
# some crazy Ryzom rendering, this merely shows how you would hook in the
# default Django rendering on external views that you include and do not want
# to fork: you can just override the default template with @html.template
# instead of by creating html templates.
class ExampleFormView(generic.FormView):
    template_name = 'form.html'
    form_class = ExampleForm

    def form_valid(self, form):
        # we don't have a success url, render again on form_valid
        return super().get(self.request)


urlpatterns = [path('', ExampleFormView.as_view())]



from django.apps import AppConfig
from django.utils.module_loading import autodiscover_modules


class BaseConfig(AppConfig):
    name = 'ryzom_django'

    def ready(self):
        autodiscover_modules('html')
        super().ready()



from django.conf import settings
from django.db import models
from django.utils import timezone

from ryzom_django_channels.pubsub import Publishable, publish


class Room(Publishable, models.Model):
    name = models.CharField(max_length=255, unique=True)

    @publish
    def rooms(cls, user):
        return cls.objects.all()


class Message(Publishable, models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        blank=True,
        null=True)
    message = models.CharField(max_length=1024)
    created = models.DateTimeField(default=timezone.now)
    room = models.ForeignKey(Room, on_delete=models.CASCADE)

    class Project:  # app_label - for app-specific model options
        ryzom = True

    def __str__(self):
        return self.message

    @publish
    def messages(cls, user):
        return cls.objects.all()



#! /usr/bin/env python

import ast
import inspect
import textwrap
import re

from . import formater


def scope(func):
    func.scope = True
    return func

class JSError(Exception):
    pass

class JS(object):

    name_map = {
        'self'   : 'this',
        'True'   : 'true',
        'False'  : 'false',
        'None'  : 'null',

        'int' : '_int',
        'float' : '_float',

        'super' : '_super',
        'print' : 'console.log',
        'undefined': 'undefined',

        # ideally we should check, that this name is available:
        'py_builtins' : '___py_hard_to_collide',
    }

    builtin = set([
        'NotImplementedError',
        'ZeroDivisionError',
        'AssertionError',
        'AttributeError',
        'RuntimeError',
        'ImportError',
        'TypeError',
        'ValueError',
        'NameError',
        'IndexError',
        'KeyError',
        'StopIteration',

        '_int',
        '_float',
        'max',
        'min',
        'sum',
    ])

    bool_op = {
        'And'    : ' && ',
        'Or'     : ' || ',
    }

    unary_op = {
        'Invert' : '~',
        'Not'    : '!',
        'UAdd'   : '+',
        'USub'   : '-',
    }

    binary_op = {
        'Add'    : '+',
        'Sub'    : '-',
        'Mult'   : '*',
        'Div'    : '/',
        'Mod'    : '%',
        'LShift' : '<<',
        'RShift' : '>>',
        'BitOr'  : '|',
        'BitXor' : '^',
        'BitAnd' : '&',
    }

    comparison_op = {
            'Eq'    : " == ",
            'NotEq' : " != ",
            'Lt'    : " < ",
            'LtE'   : " <= ",
            'Gt'    : " > ",
            'GtE'   : " >= ",
            'Is'    : " === ",
            'IsNot' : "is not", # Not implemented yet
        }

    funcs = {
            'range' : 'visit_Range',
            'len': 'visit_Len',
    }

    def __init__(self, context=None):
        self.__formater = formater.Formater()
        self.write = self.__formater.write
        self.read = self.__formater.read
        self.indent = self.__formater.indent
        self.dedent = self.__formater.dedent
        self.dummy = 0
        self.classes = ['dict', 'list', 'tuple']
        # This is the name of the class that we are currently in:
        self._class_name = None

        # This lists all variables in the local scope:
        self._scope = []
        #All calls to names within _class_names will be preceded by 'new'
        self._class_names = set()
        self._classes = {}
        self._functions = {}
        self.in_str = False

        if context:
            _copy_context = context.copy()
            for key in _copy_context.keys():
                if callable(_copy_context[key]):
                    value = context.pop(key)
                    source = inspect.getsource(value)
                    source = textwrap.dedent(source)
                    tree = ast.parse(source)
                    sub = JS()
                    sub.visit(tree)
                    self.write(sub.read())

        self._context = context or {}

    def new_dummy(self):
        dummy = "__dummy%d__" % self.dummy
        self.dummy += 1
        return dummy

    def name(self, node):
        return node.__class__.__name__

    def get_bool_op(self, node):
        return self.bool_op[node.op.__class__.__name__]

    def get_unary_op(self, node):
        return self.unary_op[node.op.__class__.__name__]

    def get_binary_op(self, node):
        return self.binary_op[node.op.__class__.__name__]

    def get_comparison_op(self, node):
        return self.comparison_op[node.__class__.__name__]

    def visit(self, node, scope=None):
        try:
            visitor = getattr(self, 'visit_' + self.name(node))
        except AttributeError:
            raise JSError("syntax not supported (%s)" % node)

        if hasattr(visitor, 'statement'):
            return visitor(node, scope)
        else:
            return visitor(node)

    def visit_Module(self, node):
        for stmt in node.body:
            self.visit(stmt)

    def visit_JoinedStr(self, node):
        result = '`'
        self.in_str = True
        for value in node.values:
            result += self.visit(value)
        self.in_str = False
        result += '`'
        return result

    def visit_FormattedValue(self, node):
        return '${%s}' % self.visit(node.value)

    def visit_AsyncFunctionDef(self, node):
        self.is_async = True
        self.visit_FunctionDef(node)

    def visit_Await(self, node):
        return f'await {self.visit(node.value)}'

    @scope
    def visit_FunctionDef(self, node):
        is_static = False
        is_javascript = False
        if node.decorator_list:
            if len(node.decorator_list) == 1 and \
                    isinstance(node.decorator_list[0], ast.Name) and \
                    node.decorator_list[0].id == "JavaScript":
                is_javascript = True # this is our own decorator
            elif self._class_name and \
                    len(node.decorator_list) == 1 and \
                    isinstance(node.decorator_list[0], ast.Name) and \
                    node.decorator_list[0].id == "staticmethod":
                is_static = True
            else:
                raise JSError("decorators are not supported")

        # XXX: disable $def for now, because it doesn't work in IE:
        if self._class_name:
        #if 1:
            if node.args.vararg is not None:
                raise JSError("star arguments are not supported")

            if node.args.kwarg is not None:
                raise JSError("keyword arguments are not supported")

            if node.decorator_list and not is_static and not is_javascript:
                raise JSError("decorators are not supported")

            defaults = [None]*(len(node.args.args) - len(node.args.defaults)) + node.args.defaults

            js_args = []
            js_defaults = []
            self._scope = [arg for arg in node.args.args]

            for arg, default in zip(node.args.args, defaults):
                if not isinstance(arg, ast.arg):
                    raise JSError("tuples in argument list are not supported")

                js_args.append(arg.arg)

                if default is not None:
                    js_defaults.append("%(id)s = typeof(%(id)s) != 'undefined' ? %(id)s : %(def)s;\n" % { 'id': arg.arg, 'def': self.visit(default) })

            if self._class_name:
                prep = "%s.prototype.%s = function(" % \
                        (self._class_name, node.name)
                if not is_static:
                    if not (js_args[0] == "self"):
                        raise NotImplementedError("The first argument must be 'self'.")
                    del js_args[0]
            else:
                prep = "function %s(" % node.name
            self.write(prep + ", ".join(js_args) + ") {")
            self.indent()
            for stmt in js_defaults:
                self.write(stmt)
            for stmt in node.body:
                self.visit(stmt)
            self.dedent()
            self.write('}')

            #If method is static, we also add it directly to the class
            if is_static:
                self.write("%s.%s = %s.prototype.%s;" % \
                        (self._class_name, node.name, self._class_name, node.name))
            #Otherwise, we wrap it to take 'self' into account
            else:
                func_name = node.name
                self.write("%s.%s = function() {" % (self._class_name, func_name))
                self.write("    %s.prototype.%s.apply(arguments[0],Array.slice(arguments,1));"% (self._class_name, func_name))
                self.write("}")

            self._scope = []
        else:
            _args = [arg for arg in node.args.args if arg.arg != 'self']
            defaults = [None]*(len(_args) - len(node.args.defaults)) + node.args.defaults

            args = []
            defaults2 = {}
            for arg, default in zip(_args, defaults):
                if arg.arg == 'self':
                    continue
                if not isinstance(arg, ast.arg):
                    raise JSError("tuples in argument list are not supported")
                if default:
                    args.append(f'{arg.arg} = {self.visit(default)}')
                else:
                    args.append(arg.arg)
            args = ", ".join(args)
            func_name = node.name
            if func_name == '__init__':
                func_name = 'constructor'
            if node.args.args and node.args.args[0].arg == 'self':
                prep = "%s(%s) {" % (func_name, args)
            else:
                prep = "function %s(%s) {" % (func_name, args)
            if getattr(self, 'is_async', False):
                prep = 'async ' + prep
                self.is_async = False
            self.write(prep)
            self._scope = [arg.arg for arg in node.args.args]
            self.indent()
            for stmt in node.body:
                self.visit(stmt)
            self.dedent()
            self.write("}")

    @scope
    def visit_ClassDef(self, node):
        if node.bases:
            self.write(f"class {node.name} extends {node.bases[0].id}" + " {")
        else:
            self.write(f"class {node.name}" + " {")
        self.indent()

        for stmt in node.body:
            if isinstance(stmt, ast.Assign):
                value = self.visit(stmt.value)
                for t in stmt.targets:
                    var = self.visit(t)
                    self.write("%s = %s;" % (var, value))
            else:
                self.visit(stmt)

        self.dedent()
        self.write("}")

    def visit_Return(self, node):
        if node.value is not None:
            self.write("return %s;" % self.visit(node.value))
        else:
            self.write("return;")

    def visit_Delete(self, node):
        return node

    @scope
    def visit_Assign(self, node):
        assert len(node.targets) == 1
        target = node.targets[0]
        value = self.visit(node.value)
        var = self.visit(target)
        if isinstance(target, ast.Name):
            if not (var in self._scope):
                self._scope.append(var)
                declare = "var "
            else:
                declare = ""
            self.write("%s%s = %s;" % (declare, var, value))
        elif isinstance(target, ast.Attribute):
            js = self.write("%s.%s = %s;" % (self.visit(target.value), str(target.attr), value))
        elif isinstance(target, ast.Subscript):
            js = self.write(f"{self.visit(target)} = {value};")
        else:
            raise JSError("Unsupported assignment type")

    def visit_AugAssign(self, node):
        # TODO: Make sure that all the logic in Assign also works in AugAssign
        target = self.visit(node.target)
        value = self.visit(node.value)

        if isinstance(node.op, ast.Pow):
            self.write("%s = Math.pow(%s, %s);" % (target, target, value))
        if isinstance(node.op, ast.FloorDiv):
            self.write("%s = Math.floor((%s)/(%s));" % (target, target, value))

        self.write("%s %s= %s;" % (target, self.get_binary_op(node), value))

    @scope
    def visit_For(self, node):
        if not isinstance(node.target, ast.Name):
            raise JSError("argument decomposition in 'for' loop is not supported")

        for_target = self.visit(node.target)
        for_iter = self.visit(node.iter)

        self.write(f"for (const {for_target} of {for_iter}) {{")
        self.indent()
        for stmt in node.body:
            self.visit(stmt)
        self.dedent()
        self.write("}")

        if node.orelse:
            self.write("if (%s) {" % orelse_dummy)
            self.indent()
            for stmt in node.orelse:
                self.visit(stmt)
            self.dedent()
            self.write("}")

    @scope
    def visit_While(self, node):

        if not node.orelse:
            self.write("while (%s) {" % self.visit(node.test))
        else:
            orelse_dummy = self.new_dummy()

            self.write("var %s = false;" % orelse_dummy)
            self.write("while (1) {");
            self.write("    if (!(%s)) {" % self.visit(node.test))
            self.write("        %s = true;" % orelse_dummy)
            self.write("        break;")
            self.write("    }")
        self.indent()
        for stmt in node.body:
            self.visit(stmt)
        self.dedent()

        self.write("}")

        if node.orelse:
            self.write("if (%s) {" % orelse_dummy)
            self.indent()
            for stmt in node.orelse:
                self.visit(stmt)
            self.dedent()
            self.write("}")

    @scope
    def visit_If(self, node):
        self.write("if (%s) {" % self.visit(node.test))
        self.indent()
        for stmt in node.body:
            self.visit(stmt)
        self.dedent()
        if node.orelse:
            self.write("} else {")
            self.indent()
            for stmt in node.orelse:
                self.visit(stmt)
            self.dedent()
        self.write("}")

    @scope
    def _visit_With(self, node):
        pass

    @scope
    def _visit_Raise(self, node):
        pass

    @scope
    def _visit_TryExcept(self, node):
        pass

    @scope
    def _visit_TryFinally(self, node):
        pass

    def visit_Assert(self, node):
        test = self.visit(node.test)

        if node.msg is not None:
            self.write("assert(%s, %s);" % (test, self.visit(node.msg)))
        else:
            self.write("assert(%s);" % test)

    def _visit_Import(self, node):
        pass

    def _visit_ImportFrom(self, node):
        pass

    def _visit_Exec(self, node):
        pass

    def visit_Global(self, node):
        self._scope.extend(node.names)

    def visit_Expr(self, node):
        self.write(self.visit(node.value) + ";")

    def visit_Pass(self, node):
        self.write("/* pass */")

    def visit_Break(self, node):
        self.write("break;")

    def visit_Continue(self, node):
        self.write("continue;")

    def visit_arguments(self, node):
        return ", ".join([self.visit(arg) for arg in node.args])

    def visit_arg(self, node):
        return node.arg

    def visit_Lambda(self, node):
        args = self.visit(node.args)
        body = self.visit(node.body)
        formater = self.__formater
        indent = formater._Formater__indent_string * (formater._Formater__indentation + 1)
        indent2 = formater._Formater__indent_string * formater._Formater__indentation
        return "\n%s(%s) => {return %s}\n%s" % (indent, args, body, indent2)

    def visit_BoolOp(self, node):
        op = self.get_bool_op(node).join([ "%s" % self.visit(val) for val in node.values ])
        return f'({op})'

    def visit_UnaryOp(self, node):
        return "%s(%s)" % (self.get_unary_op(node), self.visit(node.operand))

    def visit_BinOp(self, node):
        if isinstance(node.op, ast.Mod) and isinstance(node.left, ast.Str):
            left = self.visit(node.left)
            if isinstance(node.right, (ast.Tuple, ast.List)):
                right = self.visit(node.right)
                return "vsprintf(js(%s), js(%s))" % (left, right)
            else:
                right = self.visit(node.right)
                return "sprintf(js(%s), %s)" % (left, right)
        left = self.visit(node.left)
        right = self.visit(node.right)

        if isinstance(node.op, ast.Pow):
            return "Math.pow(%s, %s)" % (left, right)
        if isinstance(node.op, ast.FloorDiv):
            return "Math.floor((%s)/(%s))" % (left, right)

        return "(%s %s %s)" % (left, self.get_binary_op(node), right)

    def visit_Compare(self, node):
        assert len(node.ops) == 1
        assert len(node.comparators) == 1
        op = node.ops[0]
        comp = node.comparators[0]
        if isinstance(op, ast.In):
            return "%s.includes(%s)" % (
                    self.visit(comp),
                    self.visit(node.left),
                    )
        elif isinstance(op, ast.NotIn):
            return "!(%s.includes(%s))" % (
                    self.visit(comp),
                    self.visit(node.left),
                    )
        else:
            return "%s %s %s" % (self.visit(node.left),
                    self.get_comparison_op(op),
                    self.visit(comp)
                    )

    def visit_Name(self, node):
        id = node.id
        try:
            id = self.name_map[id]
        except KeyError:
            pass

        if id in self.builtin:
            id = "py_builtins." + id;

        #~ if id in self._classes:
            #~ id = '_' + id;
        elif self._context and id in self._context:
            value = self._context[id]
            if isinstance(value, str):
                return "\"%s\"" % value
            else:
                return "%s" % value

        return id

    def visit_Constant(self, node):
        if isinstance(node.value, str):
            return self.visit_Str(node)
        elif isinstance(node.value, bool):
            return 'true' if node.value else 'false'
        elif node.value is None:
            return 'null'
        else:
            return self.visit_Num(node)

    def visit_Num(self, node):
        return node.n

    def visit_Str(self, node):
        # Uses the Python builtin repr() of a string and the strip string type
        # from it. This is to ensure Javascriptness, even when they use things
        # like b"\\x00" or u"\\u0000".
        if not self.in_str:
            return "%s" % repr(node.s).lstrip("urb")
        return node.s

    def visit_Call(self, node):
        func = self.visit(node.func)
        #~ if func in self._class_names:
            #~ func = 'new '+func

        if getattr(node.func, 'id', None) == 'type':
            return "typeof %s" % self.visit(node.args[0])
        elif getattr(node.func, 'id', None) == 'bool':
            return "new Boolean(%s)" % self.visit(node.args[0])
        elif node.keywords:
            keywords = []
            for kw in node.keywords:
                keywords.append("%s: %s" % (kw.arg, self.visit(kw.value)))
            keywords = "{" + ", ".join(keywords) + "}"
            js_args = ",".join([ self.visit(arg) for arg in node.args ])
            return "%s.args([%s], %s)" % (func, js_args,
                    keywords)
        else:
            if getattr(node, 'starargs', None) is not None:
                raise JSError("star arguments are not supported")

            if getattr(node, 'kwargs', None) is not None:
                raise JSError("keyword arguments are not supported")

            try:
                js_args = ",".join([str(self.visit(arg)) for arg in node.args ])
            except:
                raise

            if func in self.funcs.keys():
                method = getattr(self, self.funcs[func])
                return method(js_args)

            if func.endswith('.endswith'):
                func = func[:-4] + 'With'

            return "%s(%s)" % (func, js_args)

    def visit_Raise(self, node):
        assert node.inst is None
        assert node.tback is None
        self.write("throw %s;" % self.visit(node.type))

    def visit_Print(self, node):
        assert node.dest is None
        assert node.nl
        values = [self.visit(v) for v in node.values]
        values = ", ".join(values)
        self.write("py_builtins.print(%s);" % values)

    def visit_Attribute(self, node):
        if (
            isinstance(node.value, ast.Name)
            and node.value.id in self._context
            and hasattr(self._context[node.value.id], node.attr)
        ):
            obj = self._context[node.value.id]
            value = getattr(obj, node.attr)
            if isinstance(value, str):
                return f'"{value}"'
            elif callable(value):
                base_name = ''
                try:
                    for cls in inspect.getmro(value.__self__.__class__):
                        if value.__name__ in cls.__dict__:
                            base_name = cls.__name__
                except AttributeError:
                    pass
                if not base_name:
                    base_name = getattr(obj, '__name__', type(obj).__name__)
                name = base_name + f'_{node.attr}'
                if name not in self._functions:
                    self._functions[name] = value
                return name
            elif str(value) in self.name_map:
                return self.name_map[str(value)]
            elif isinstance(value, (int, float)):
                return f'{value}'
            else:
                return value

        value = self.visit(node.value)

        if value == 'new':
            return "%s %s" % (value, node.attr)

        if not isinstance(value, str):
            value = getattr(value, node.attr)
            if isinstance(value, str):
                return f'"{value}"'
            return value
        return "%s.%s" % (self.visit(node.value), node.attr)

    def visit_Tuple(self, node):
        els = [self.visit(e) for e in node.elts]
        return "[%s]" % (", ".join(els))

    def visit_Dict(self, node):
        els = []
        for k, v in zip(node.keys, node.values):
            els.append('%s: %s' % (self.visit(k), self.visit(v)))

        if len(els) > 1:
            indent_str = self.__formater._Formater__indent_string
            indent_lvl = self.__formater._Formater__indentation
            indent = indent_str * indent_lvl
            indent2 = indent + indent_str
            return "{\n%s%s\n%s}" % (indent2, f",\n{indent2}".join(els), indent)
        return "{%s}" % (", ".join(els))

    def visit_List(self, node):
        els = [str(self.visit(e)) for e in node.elts]
        return "[%s]" % (", ".join(els))

    def visit_Slice(self, node):
        if node.lower and node.upper and node.step:
            return "slice(%s, %s, %s)" % (self.visit(node.lower),
                    self.visit(node.upper), self.visit(node.step))
        if node.lower and node.upper:
            return "slice(%s, %s)" % (self.visit(node.lower),
                    self.visit(node.upper))
        if node.upper and not node.step:
            return "slice(%s)" % (self.visit(node.upper))
        if node.lower and not node.step:
            return "slice(%s, null)" % (self.visit(node.lower))
        if not node.lower and not node.upper and not node.step:
            return "slice(null)"
        raise NotImplementedError("Slice")

    def visit_Subscript(self, node):
        return "%s[%s]" % (self.visit(node.value), self.visit(node.slice))

    def visit_Index(self, node):
        return self.visit(node.value)

    #specifics
    def visit_Range(self, args):
        return f"[...Array({args}).keys()]"

    def visit_Len(self, args):
        return f"{args}.length"


def convert_py2js(s, context=None):
    """
    Takes Python code as a string 's' and converts this to JavaScript.

    Example:

    >>> convert_py2js("x[3:]")
    'x.__getitem__(slice(3, null));'

    """
    t = ast.parse(textwrap.dedent(s))
    v = JS(context)
    v.visit(t)
    js = v.read()
    return js


def transpile(obj_or_src, **context):
    """
    Transpile a Python object or source code into javascript.
    """
    if isinstance(obj_or_src, str):
        src = obj_or_src
    else:
        src = inspect.getsource(obj_or_src)
    return convert_py2js(src, context)


def transpile_body(obj, **context):
    src = inspect.getsource(obj)
    src = '\n'.join(src.split('\n')[1:])
    src = textwrap.dedent(src)
    return transpile(src, **context)


def transpile_class(cls, superclass=None, newname=None, **context):
    src = inspect.getsource(cls)
    lines = src.split('\n')
    if newname:
        lines[0] = re.sub('class ([^:])*:', f'class {newname}:', lines[0])
    if superclass:
        lines[0] = re.sub('[()]*:', f'({superclass}):', lines[0])
    src = '\n'.join(lines)
    src = textwrap.dedent(src)
    return transpile(src, **context)


def transpile_function(func, newname=None, **context):
    src = inspect.getsource(func)
    src = textwrap.dedent(src)
    lines = src.split('\n')
    if newname:
        lines[0] = re.sub('def ([^(])*', f'def {newname}', lines[0])
    src = '\n'.join(lines)
    return transpile(src, **context)



# Generated by Django 3.1.7 on 2021-03-22 16:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ryzom_django_channels', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='publication',
            name='template_class',
        ),
        migrations.RemoveField(
            model_name='publication',
            name='template_module',
        ),
    ]



from django import forms, http
from django.urls import path, reverse
from django.views import generic

import py2js
from py2js.renderer import JS
from ryzom_django_channels.components import (
    ReactiveComponentMixin,
    SubscribeComponentMixin,
    model_template
)
from ryzom_django_channels.views import ReactiveMixin, register
from ryzom_django_mdc.html import *

from .models import Message, Room


class AjaxFormMixin:
    async def on_form_submit(event):
        event.preventDefault()

        form = event.target

        await fetch(form.action, {
            'method': form.method,
            'body': new.FormData(form)
        }).then(
            lambda response : print(response)
        )

        form.reset()

    def py2js(self):
        form = getElementByUuid(self.id)
        form.addEventListener('submit', self.on_form_submit)


class MessageFormComponent(AjaxFormMixin, Form):
    def __init__(self, *content, view, form, **context):
        super().__init__(
            Div(
                form,
                MDCButton(form.submit_label or 'submit'),
                style='display:flex; flex-flow: row nowrap; align-items: baseline;'),
            CSRFInput(view.request),
            method='POST',
            **context)


class DeleteButton(Component):
    tag = 'delete-button'

    def __init__(self, *content, **attrs):
        content = content or (MDCButtonOutlined(MDCIcon('delete')),)
        super().__init__(*content, **attrs)

    class HTMLElement:
        def connectedCallback(self):
            this.addEventListener('click', this.delete.bind(this))

        async def delete(self, event):
            csrf = document.querySelector('[name="csrfmiddlewaretoken"]')

            print('url', this.attributes['delete-url'].value)
            await fetch(this.attributes['delete-url'].value, {
                method: 'delete',
                headers: {'X-CSRFTOKEN': csrf.value},
                redirect: 'manual',
            }).then(
                lambda response: print(response)
            )


@model_template('message-item')
class MessageItem(MDCListItem):
    def __init__(self, obj):
        self.obj = obj

        username = obj.user.username if obj.user else 'Anonymous'

        super().__init__(
            Span(username, ' says: ', obj.message),
            DeleteButton(
                delete_url=reverse('message_delete', args=[self.obj.id]),
            ),
            id=f'message-{obj.id}',
        )


class ChatRoom(SubscribeComponentMixin, MDCList):
    publication = 'messages'
    model_template = 'message-item'

    def __init__(self, room_id):
        self.subscribe_options = dict(room_id=room_id)
        super().__init__()

    @classmethod
    def get_queryset(cls, user, qs, opts):
        room_messages = qs.filter(room__name=opts['room_id']).order_by('created')
        count = room_messages.count()
        start = max(count - 10, 0)
        return room_messages[start:count]


@model_template('room-item')
class RoomItem(MDCListItem):
    def __init__(self, room):
        super().__init__(
            room.name,
            id=f'room-{room.id}',
            tag='a',
            href=f'/reactive/?room={room.name}')


class RoomForm(Div):
    def __init__(self, order_by):
        super().__init__(
            Form(
                MDCTextFieldOutlined(
                    MDCTextInput('room'), 'Room Name'),
                MDCButton('go'),
                RoomList(order_by),
            )
        )


class RoomList(SubscribeComponentMixin, MDCList):
    publication = 'rooms'
    model_template = 'room-item'

    def __init__(self, order_by):
        self.subscribe_options = dict(order_by=order_by)
        super().__init__()

    @classmethod
    def get_queryset(self, user, qs, opts):
        return qs.order_by(opts['order_by'])


class Body(Body):
    def __init__(self, *content, **context):
        super().__init__(
            Style(
                'form div, form .mdc-text-field, .mdc-list-item__text {width: 100%;}'
                + '.mdc-list-item__text {display: flex; justify-content: space-between;',
            ),
            *content,
        )


class ReactiveTitle(ReactiveComponentMixin, H1):
    register = 'page_title'

    def __init__(self, room):
        super().__init__(f'Test - {room}')


@template('home')
class Home(Html):
    scripts = ['/static/ryzom.js']
    body_class = Body

    def to_html(self, *content, view, form, **context):
        current_room_name = view.request.GET.get('room', 'general')
        current_room = Room.objects.filter(name=current_room_name).first()
        message_count = current_room.message_set.count() if current_room else 0

        head, body = content

        body.addchildren([
            ReactiveTitle(f'{current_room_name} - {message_count} messages'),
            A('test forms', href='form/'),
            Div(
                Div(
                    RoomForm(
                        view.request.GET.get('order_by', 'name')),
                    style='min-width: 20%'),
                Div(
                    ChatRoom(current_room_name),
                    message_form=MessageFormComponent(
                        view=view,
                        form=form,
                        style='width:100%'),
                    style='flex-grow: 1; height: 100%;'),
                style='display:flex; flex-flow: row wrap;'),
            Script(view.get_token()),
            Script('mdc.autoInit();'),
        ])

        return super().to_html(
            head,
            body,
        )


class MessageForm(forms.ModelForm):
    submit_label = 'Send'

    message = forms.CharField(
        required=True,
        help_text='type your message here')

    class Meta:
        model = Message
        fields = ['message']


class ChatView(ReactiveMixin, generic.CreateView):
    template_name = 'home'
    form_class = MessageForm

    def form_valid(self, form):
        if self.request.user.is_authenticated:
            form.instance.user = self.request.user

        room = self.request.GET.get('room', 'general')
        room_obj, _ = Room.objects.get_or_create(name=room)

        form.instance.room = room_obj
        return super().form_valid(form)

    def get_success_url(self):
        msg = self.object
        room = msg.room
        message_count = msg.room.message_set.count()
        register('page_title').update(
            ReactiveTitle(f'{msg.room.name} - {message_count} messages'))

        return reverse('home')

    @classmethod
    def as_url(cls):
        return path('', cls.as_view(), name='home')


class ChatDeleteView(generic.DeleteView):
    model = Message

    def dispatch(self, request, *args, **kwargs):
        return self.delete(request, *args, **kwargs)

    def get_success_url(self):
        msg = self.get_object()
        room = msg.room
        message_count = msg.room.message_set.count() - 1
        register('page_title').update(
            ReactiveTitle(f'{msg.room.name} - {message_count} messages'))

        if not message_count and room.name != 'general':
            room.delete()

        return reverse('home')

    @classmethod
    def as_url(cls):
        return path('message/<pk>/delete', cls.as_view(), name='message_delete')


urlpatterns = [
    ChatView.as_url(),
    ChatDeleteView.as_url(),
]



import os

from django.core.management.base import BaseCommand, CommandError

from ryzom_django import bundle


class Command(BaseCommand):
    help = 'Write JS & CSS bundles to ryzom_django/static/bundle.*'

    def handle(self, *args, **options):
        static_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                '..',
                '..',
                'static',
            )
        )

        if not os.path.exists(static_path):
            os.makedirs(static_path)

        with open(f'{static_path}/bundle.js', 'w+') as f:
            f.write(bundle.js())

        with open(f'{static_path}/bundle.css', 'w+') as f:
            f.write(bundle.css())

        self.stdout.write(self.style.SUCCESS(f'Successfully wrote {static_path}/bundle.*'))
        self.stdout.write('Do not forget to collectstatic!')



from py2js import transpile


class js_renderer(object):
    """
    Decorator that you can use to convert methods to JavaScript.

    For example this code::

        @JavaScript
        class TestClass(object):
            def __init__(self):
                alert('TestClass created')
                self.reset()

            def reset(self):
                self.value = 0

            def inc(self):
                alert(self.value)
                self.value += 1

        print str(TestClass)

    prints::

        function TestClass() {
            return new _TestClass();
        }
        function _TestClass() {
            this.__init__();
        }
        _TestClass.__name__ = 'TestClass'
        _TestClass.prototype.__class__ = _TestClass
        _TestClass.prototype.__init__ = function() {
            alert("TestClass created");
            this.reset();
        }
        _TestClass.prototype.reset = function() {
            this.value = 0;
        }
        _TestClass.prototype.inc = function() {
            alert(this.value);
            this.value += 1;
        }

    Alternatively, an equivalent way is to use JavaScript() as a function:

        class TestClass(object):
            def __init__(self):
                alert('TestClass created')
                self.reset()

            def reset(self):
                self.value = 0

            def inc(self):
                alert(self.value)
                self.value += 1

        print str(JavaScript(TestClass))

    If you want to call the original function/class as Python, use the
    following syntax::

        js = JavaScript(TestClass)
        test_class = js() # Python instance of TestClass() will be created
        js_class = str(js) # A string with the JS code

    """

    def __init__(self, obj, context=None):
        self._obj = obj
        self._js = transpile(obj, **(context or {}))

    def __str__(self):
        return self._js

    def __call__(self, *args, **kwargs):
        return self._obj(*args, **kwargs)


def JS(obj, ctx=None):
    return str(js_renderer(obj, ctx))


def autoexec(js):
    if js:
        return '\n(' + str(js)[:-1] + ')();\n'
    return ''



'''
This file defines the models needed for ryzom pub/sub system.
They're not intended to be used by end-user.
'''
import importlib
import secrets
import uuid

from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.postgres.fields import ArrayField, JSONField
from django.db import models
from django.db.models import JSONField
from django.utils import timezone


class Client(models.Model):
    '''
    Clients are the representation of connected Clients
    over websockets. It stores the channel name of a single
    client to communicate over the channel layer
    The user field is not used for now but in a near future
    it will be used to store the user using this channel
    once it's connected
    '''
    token = models.CharField(default=secrets.token_urlsafe,
                             max_length=255, unique=True)
    channel = models.CharField(max_length=255)
    user = models.ForeignKey(
                settings.AUTH_USER_MODEL,
                models.SET_NULL,
                blank=True,
                null=True
           )
    created = models.DateTimeField(default=timezone.now)


class Registration(models.Model):
    name = models.CharField(max_length=255)
    client = models.ForeignKey(Client, models.CASCADE, blank=True, null=True)
    subscriber_id = models.CharField(max_length=255)
    subscriber_parent = models.CharField(max_length=255)

    class Meta:
        unique_together = [('name', 'client')]


class Publication(models.Model):
    '''
    Publications model is used to store the apps publications
    Each publication should have a unique name and define
    the component used as template for the publicated model.
    One can publish a model multiple time with varying templates
    or query.
    The query is a JSON field containing informations on what and
    how to publish about the model concerned, such as
    order_by, limit, offset, filters and more soon
    '''
    name = models.CharField(max_length=255, unique=True)
    model_module = models.CharField(max_length=255)
    model_class = models.CharField(max_length=255)

    @property
    def publish_function(self):
        model = getattr(
            importlib.import_module(self.model_module),
            self.model_class
        )
        return getattr(model, self.name)


class Subscription(models.Model):
    '''
    A subscription is an object representing the relation between
    a client and a publication. It also stores the id of the component
    that subscribes to a given publication, and the queryset
    computed from that publication query. This queryset is computed
    per-subscription to permit user specific sets
    After being instanciated, a subscription must be initialized by
    it's init() method so that it fills the component asking for it
    by its content via ryzom.ddp send_insert.
    '''
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    client = models.ForeignKey(Client, models.CASCADE, blank=True, null=True)
    publication = models.ForeignKey(Publication, models.CASCADE)
    subscriber_id = models.CharField(max_length=255)
    subscriber_module = models.CharField(max_length=255)
    subscriber_class = models.CharField(max_length=255)
    queryset = ArrayField(models.IntegerField(), default=list)
    options = JSONField(blank=True, null=True)

    @property
    def subscriber(self):
        subscriber_mod = importlib.import_module(self.subscriber_module)
        return getattr(subscriber_mod, self.subscriber_class)

    def get_queryset(self, opts=None):  # noqa: C901
        '''
        This method computes the publication query and create/update the
        queryset for the current subscription.
        It supports somme special variables such as $count and $user that
        are parsed and replaced with, respectively, the queryset.count()
        value and the current user associated with the subscription client
        More will come with special variables and function. Such as an $add
        to replace that ugly tupple i'm using for now.. to be discussed
        '''

        queryset = self.publication.publish_function(self.client.user)

        opts = opts or self.options
        queryset = self.subscriber.get_queryset(
            self.client.user, queryset, opts)

        self.options = opts
        self.queryset = list(queryset.values_list('id', flat=True))
        self.save()

        return queryset



from django.apps import AppConfig, apps
from django.db import OperationalError, ProgrammingError


class BaseConfig(AppConfig):
    '''
    Ryzom application configuration
    '''
    name = 'ryzom_django_channels'

    def ready(self):
        '''
        AppConfig.ready overloading.

        Here we import the ryzom.signals module to connect the signals handlers
        to receivers, then we clean the Clients table (zombies clients stay
        after a server reboot).

        Finally, we create the publications that were registered
        in the to_publish list of ryzom.pusub module.
        '''
        import ryzom_django_channels.signals
        from ryzom_django_channels.pubsub import Publishable

        try:
            Client = self.get_model('Client')
            Client.objects.all().delete()

            for model in apps.get_models():
                if issubclass(model, Publishable):
                    model.publish()

        except (OperationalError, ProgrammingError):
            pass

        super().ready()



'''
Defines the HTTP routes to static files and ryzom index.
This will be subject to change in a near future, when implementing SSR
'''
from django.conf import settings
from django.urls import path

from . import views

urlpatterns = [
    path(f'{settings.RYZOM_APP}', views.index, name='index'),
    path('ws/ddp', views.index, name='index'),
    path('<path:url>', views.index, name='index')
]



'''
Functions to communicate DDP messages to the channel layer.
'''
from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer


def send_insert(sub, model, tmpl, id):
    '''
    Send insert message.
    Function used to send a DDP message to a specific client
    via the channel layer.
    Uses the template class associated with a publication
    to create a new instance of a component attached to a
    model that was inserted, updated or removed
    Essentially called by post_save and post_delete signal handlers

    :param Subscriptions sub: The Subscription holding the connection \
            information
    :param Publishable model: The class of the model to insert
    :param Component tmpl: The component subclass that templates \
            the model instance
    :param int id: The id of the model to insert
    '''
    if sub.client is None or sub.client.channel == '':
        return

    tmpl_instance = tmpl(model.objects.get(id=id))
    tmpl_instance.parent = sub.subscriber_id
    tmpl_instance.position = sub.queryset.index(id)
    data = {
        'type': 'handle.ddp',
        'params': {
            'type': 'inserted',
            'instance': tmpl_instance.to_obj()
        }
    }
    channel = get_channel_layer()
    async_to_sync(channel.send)(sub.client.channel, data)


def send_change(sub, model, tmpl, id):
    '''
    Send change message.
    Function used to send a DDP message to a specific client
    via the channel layer.
    Uses the template class associated with a publication
    to create a new instance of a component attached to a
    model that was updated
    Essentially called by post_save and post_delete signal handlers

    :param Subscriptions sub: The Subscription holding the connection \
            information
    :param Publishable model: The class of the model to change
    :param Component tmpl: The component subclass that templates \
            the model instance
    :param int id: The id of the model to change
    '''
    if sub.client is None or sub.client.channel == '':
        return

    tmpl_instance = tmpl(model.objects.get(id=id))
    tmpl_instance.parent = sub.subscriber_id
    tmpl_instance.position = sub.queryset.index(id)
    data = {
        'type': 'handle.ddp',
        'params': {
            'type': 'changed',
            'instance': tmpl_instance.to_obj()
        }
    }
    channel = get_channel_layer()
    async_to_sync(channel.send)(sub.client.channel, data)


def send_remove(sub, model, tmpl, id):
    '''
    Send remove message.
    Function used to send a DDP message to a specific client
    via the channel layer.
    Uses the template class associated with a publication
    to create a new instance of a component attached to a
    model that was removed, in order to get the computed id
    and send the computed id to the client.
    Essentially called by post_save and post_delete signal handlers

    :param Subscriptions sub: The Subscription holding the connection \
            information
    :param Publishable model: The class of the model to remove
    :param Component tmpl: The component subclass that templates \
            the model instance
    :param int id: The id of the model to remove
    '''
    if sub.client is None or sub.client.channel == '':
        return

    tmp = model()
    tmp.id = id
    tmpl_instance = tmpl(tmp)
    data = {
        'type': 'handle.ddp',
        'params': {
            'type': 'removed',
            'id': tmpl_instance.id,
            'parent': sub.subscriber_id
        }
    }
    channel = get_channel_layer()
    async_to_sync(channel.send)(sub.client.channel, data)



class Publishable:
    '''
    The publishable class is meant to be inherited from
    by end user models. It defines a publish class method
    that's used to create a publication associated with the
    model inherited from it.
    '''

    @classmethod
    def publish(cls):
        for method in cls.__dict__.values():
            if getattr(method, '__publication__', False):
                cls.create_publication(getattr(method, '__name__'))

    @classmethod
    def create_publication(cls, name):
        from ryzom_django_channels.models import Publication

        model_params = dict(
            model_module=cls.__module__,
            model_class=cls.__name__,
        )
        if pub := Publication.objects.filter(name=name).first():
            changed = False
            for k, v in model_params.items():
                if getattr(pub, k, None) != v:
                    setattr(pub, k, v)
                    changed = True
            if changed:
                pub.save()
        else:
            Publication.objects.create(name=name, **model_params)


def publish(func):
    @classmethod
    def wrapper(*args):
        return func(*args)
    wrapper.__publication__ = True
    wrapper.__name__ = func.__name__
    return wrapper



import ast
import inspect
import importlib
import py2js
from py2js.transpiler import JS
import re
import textwrap


AUTOCOMPILE = (
    'onclick',
    'onmouseover',
    'onsubmit',
    'onchange',
    'oninput',
)


def webcomponent(value):
    return [
        py2js.transpile_class(
            value.HTMLElement,
            superclass='HTMLElement',
            newname=value.__name__,
        ),
        f'window.customElements.define("{value.tag}", {value.__name__});',
    ]



def _functions(transpiler, done, **context):
    out = []
    for name, func in transpiler._functions.items():
        if func in done:
            continue
        func_src = textwrap.dedent(inspect.getsource(func))
        func_ast = ast.parse(func_src)
        func_ast.body[0].name = name
        func_js = JS()
        func_js._context = context
        func_js.visit(func_ast)
        out.append(func_js.read())
        if func_js._functions:
            out += _functions(func_js, done, **context)
        done.append(func)
    return out


def functions(value, done):
    out = []
    tree = ast.parse(textwrap.dedent(inspect.getsource(value.py2js)))
    transpiler = JS()
    transpiler._context = dict(self=value)
    transpiler.visit(tree)
    out += _functions(transpiler, done, self=value)
    return out


def methods(value, done_funcs):
    out = []

    for name in AUTOCOMPILE:
        method = getattr(value, name, None)
        if not method:
            continue
        src = textwrap.dedent(inspect.getsource(method))
        src = re.sub(r'^def ', f'def {value.__name__}_', src)
        tree = ast.parse(src)
        transpiler = JS()
        transpiler._context = dict(self=value)
        transpiler.visit(tree)
        out += [transpiler.read()]
        out += _functions(transpiler, done_funcs, self=value)

    return out


def bundle(*modules):
    out = []
    done = []
    done_funcs = []
    for module in modules:
        mod = importlib.import_module(module)
        for key, value in mod.__dict__.items():
            if value in done:
                continue
            if hasattr(value, 'HTMLElement'):
                out += webcomponent(value)
            if callable(getattr(value, 'py2js', None)):
                out += functions(value, done_funcs)
            out += methods(value, done_funcs)
            done.append(value)
    return '\n'.join(out)



'''
Defines the django signals handlers.
'''
import importlib

from django.contrib.postgres.aggregates import ArrayAgg
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from ryzom_django_channels.components import model_templates
from ryzom_django_channels.ddp import send_change, send_insert, send_remove
from ryzom_django_channels.models import Publication, Subscription
from ryzom_django_channels.pubsub import Publishable


@receiver(post_save)
def _ddp_insert_change(sender, **kwargs):
    '''
    Function to send a DDP insert/change/remove messages to the channel layer
    whenever a Publishable model's save() method is called.
    This function will update the queryset of all subscriptions
    associated with the sender model and send insert, remove or change
    message for each id that was added or removed from the old
    queryset to the new one.
    '''
    if Publishable not in sender.mro():
        return

    created = kwargs.pop('created')
    instance = kwargs.pop('instance')

    subscriptions = Subscription.objects.filter(
        publication__model_class=sender.__name__,
        publication__model_module=sender.__module__)

    for sub in subscriptions:
        template = model_templates[sub.subscriber.model_template]
        old_qs = sub.queryset

        sub.get_queryset()
        new_qs = sub.queryset

        diff = {
            'inserted': set(new_qs).difference(set(old_qs)),
            'removed': set(old_qs).difference(set(new_qs))
        }
        # if sets are the same
        if not diff['inserted'] and not diff['removed']:
            # if created and sets are the same,
            # entry has been filtered and can't be there
            if not created:
                # changed and may have moved
                # just send new instance and pos
                send_change(sub, sender, template, instance.id)

        # if sets aren't the same, then considering that only one entry
        # was added or has changed:
        # - it could have been removed if newly filtered (changed)
        # - it could have been added if no more filtered (changed)
        # - it could have been added if created and not filtered
        # - it cannot have just moved neither changed or the set
        #   would have been the same
        # - it could have been added while not created, replacing
        #   another entry because of filters, so created or not,
        #   we have to handle both added and removed entries
        else:
            # using loops for now but shouldn't be usefull as we
            # are handling only one entry, the queryset shouldn't
            # move by more that one in and/or one out
            for id in diff['removed']:
                send_remove(sub, sender, template, id)
            for id in diff['inserted']:
                send_insert(sub, sender, template, id)


@receiver(post_delete)
def _ddp_delete(sender, **kwargs):
    '''
    Function to send a DDP insert/remove messages to the channel layer
    whenever a Publishable model's delete() method is called.
    This function will update the queryset of all subscriptions
    associated with the sender model and send insert and remove
    message for each id that was added or removed from the old
    queryset to the new one.
    '''
    if Publishable not in sender.mro():
        return

    instance = kwargs.pop('instance')
    subscriptions = Subscription.objects.filter(
        publication__model_class=sender.__name__,
        publication__model_module=sender.__module__)

    for sub in subscriptions:
        template = model_templates[sub.subscriber.model_template]

        old_qs = sub.queryset

        # if instance not in queryset, no need to remove it
        # or update the queryset
        # else:
        if instance.id in old_qs:
            sub.get_queryset()
            new_qs = sub.queryset

            diff = {
                'inserted': set(new_qs).difference(set(old_qs)),
                'removed': set(old_qs).difference(set(new_qs))
            }

            for id in diff['removed']:
                send_remove(sub, sender, template, id)
            for id in diff['inserted']:
                send_insert(sub, sender, template, id)



class Methods():
    methods = {}

    @classmethod
    def add(cls, methods):
        cls.methods.update(methods)

    @classmethod
    def get(cls, name):
        return cls.methods.get(name, None)



from django.middleware.csrf import get_token
from ryzom_django.forms import widget_template
from ryzom_django.html import *
from ryzom_django_mdc.forms import (context_attrs, field_kwargs, widget_attrs,
                                    widget_context)

from ryzom_django.html import *
from ryzom_mdc.html import *
from ryzom_django import html as ryzom_django


class Html(ryzom_django.Html, Html):
    pass


@widget_template(
    'django/forms/widgets/input.html',
    'django/forms/widgets/date.html',
    'django/forms/widgets/time.html',
    'django/forms/widgets/text.html',
    'django/forms/widgets/email.html',
    'django/forms/widgets/password.html',
    'django/forms/widgets/number.html',
)
class MDCInputWidget(MDCTextFieldOutlined):
    @classmethod
    def from_boundfield(cls, bf, **attrs):
        attrs.update(widget_attrs(bf))
        return cls(
            Input(**attrs),
            label=bf.label,
            help_text=bf.help_text,
            errors=bf.errors,
        )


@widget_template('django/forms/widgets/checkbox.html')
class MDCCheckboxWidget(MDCCheckboxField):
    @classmethod
    def from_boundfield(cls, bf, **attrs):
        attrs.update(widget_attrs(bf))
        return cls(
            MDCCheckboxInput(**attrs),
            **field_kwargs(bf),
        )


@widget_template('django/forms/widgets/checkbox_select.html')
class MDCCheckboxSelectMultipleWidget(MDCCheckboxSelectField):
    @classmethod
    def from_boundfield(cls, bf, **attrs):
        context = widget_context(bf)
        choices = []
        for group, options, index in context['optgroups']:
            if group is not None:
                continue
            choices += options

        return cls(
            Label(bf.label),
            *[
                Div(
                    MDCFormField(
                        MDCCheckboxInput(**context_attrs(choice)),
                        Label(choice['label'])
                    )
                )
                for choice in choices
            ],
            **field_kwargs(bf),
        )


@widget_template('django/forms/widgets/multiwidget.html')
class MultiWidget(CList):
    def __init__(self, **context):
        super().__init__(*[
            templates[widget['template_name']](**widget)
            for widget in context['widget']['subwidgets']
        ])

    @classmethod
    def from_boundfield(cls, bf, **attrs):
        attrs.update(widget_context(bf))
        return Div(
            Label(bf.label),
            cls(**attrs),
            cls='mdc-form-field'
        )


@widget_template('django/forms/widgets/splitdatetime.html')
class SplitDateTimeWidget(MDCField):
    time_style = 'margin-bottom: 0; margin-top: 12px'
    date_style = 'margin-right: 12px; ' + time_style

    @classmethod
    def from_boundfield(cls, bf, **attrs):
        context = widget_context(bf)
        return cls(
            Label(bf.label, style='display: block'),
            MDCFormField(
                MDCTextFieldOutlined(
                    Input(**context_attrs(context['subwidgets'][0])),
                    label='Date',
                    style=cls.date_style
                ),
                MDCTextFieldOutlined(
                    Input(**context_attrs(context['subwidgets'][1])),
                    label='Time',
                    style=cls.time_style,
                ),
            ),
            **field_kwargs(bf),
        )


@widget_template('django/forms/widgets/textarea.html')
class TextareaWidget(MDCTextareaFieldOutlined):
    @classmethod
    def from_boundfield(cls, bf, **attrs):
        attrs.update(widget_attrs(bf))
        return cls(
            Textarea(
                attrs.pop('value', '') or '',
                aria_labelledby=f'id_{bf.html_name}_label',
                **attrs,
            ),
            label=bf.label,
            help_text=bf.help_text,
            errors=bf.errors,
        )


@widget_template('django/forms/widgets/file.html')
class FileInputWidget(MDCField):
    @classmethod
    def from_boundfield(cls, bf, **attrs):
        attrs.update(widget_attrs(bf))
        return cls(
            Label(bf.label),
            MDCFileField(
                Input(**attrs),
                label='Select file',
                **attrs),
            name=attrs['name']
        )


@widget_template('django/forms/widgets/select.html')
class SelectWidget(MDCField):
    @classmethod
    def from_boundfield(cls, bf, **attrs):
        context = widget_context(bf)
        attrs.update(widget_attrs(bf))
        context['label'] = bf.label
        return cls(
            MDCFormField(
                MDCSelectOutlined(**context),
            ),
            name=attrs['name']
        )


class SimpleForm(Form):
    def __init__(self, view, form):
        label = getattr(form, 'submit_label', 'submit')
        super().__init__(
            CSRFInput(view.request),
            form,
            MDCButton(label),
            method='POST')



# Generated by Django 3.1.4 on 2021-03-14 19:54

import secrets
import uuid

import django.contrib.postgres.fields
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Client',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('token', models.CharField(default=secrets.token_urlsafe, max_length=255, unique=True)),
                ('channel', models.CharField(max_length=255)),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Publication',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
                ('model_module', models.CharField(max_length=255)),
                ('model_class', models.CharField(max_length=255)),
                ('template_module', models.CharField(max_length=255)),
                ('template_class', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('subscriber_id', models.CharField(max_length=255)),
                ('subscriber_module', models.CharField(max_length=255)),
                ('subscriber_class', models.CharField(max_length=255)),
                ('queryset', django.contrib.postgres.fields.ArrayField(base_field=models.IntegerField(), default=list, size=None)),
                ('options', models.JSONField(blank=True, null=True)),
                ('client', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='ryzom_django_channels.client')),
                ('publication', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ryzom_django_channels.publication')),
            ],
        ),
        migrations.CreateModel(
            name='Registration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('subscriber_id', models.CharField(max_length=255)),
                ('subscriber_parent', models.CharField(max_length=255)),
                ('client', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='ryzom_django_channels.client')),
            ],
            options={
                'unique_together': {('name', 'client')},
            },
        ),
    ]



""" A formater module that keeps trac of indentation
"""
class Formater(object):
    """
    A very simple code formater that handles efficient concatenation and indentation of lines.
    """

    def __init__(self, indent_string="    "):
        self.__buffer = []
        self.__indentation = 0
        self.__indent_string = indent_string
        self.__indent_temp = ""
        self.__string_buffer = ""

    def dedent(self):
        """
        Subtracts one indentation level.
        """
        self.__indentation -= 1
        self.__indent_temp = self.__indent_string*self.__indentation

    def indent(self):
        """
        Adds one indentation level.
        """
        self.__indentation += 1
        self.__indent_temp = self.__indent_string*self.__indentation

    def write(self, text, indent=True, newline=True):
        """
        Writes the string text to the buffer with indentation and a newline if not specified otherwise.
        """
        if indent:
            self.__buffer.append(self.__indent_temp)
        self.__buffer.append(text)
        if newline:
            self.__buffer.append("\n")

    def read(self, size=None):
        """
        Returns a string representation of the buffer.
        """
        if size == None:
            text = self.__string_buffer + "".join(self.__buffer)
            self.__buffer = []
            self.__string_buffer = ""
            return text
        else:
            if len(self.__string_buffer) < size:
                self.__string_buffer += "".join(self.__buffer)
                self.__buffer = []
                if len(self.__string_buffer) < size:
                    text, self.__string_buffer = self.__string_buffer, ""
                    return text
                else:
                    text, self.__string_buffer = self.__string_buffer[:size], \
                        self.__string_buffer[size:]
                    return text
            else:
                text, self.__string_buffer = self.__string_buffer[:size], \
                    self.__string_buffer[size:]
                return text

